;; The arithmetic sequence, 1487, 4817, 8147, in which each of the
;; terms increases by 3330, is unusual in two ways: (i) each of the
;; three terms are prime, and, (ii) each of the 4-digit numbers are
;; permutations of one another.

;; There are no arithmetic sequences made up of three 1-, 2-, or
;; 3-digit primes, exhibiting this property, but there is one other
;; 4-digit increasing sequence.

;; What 12-digit number do you form by concatenating the three terms
;; in this sequence?

;; has to be in the set of primes which are 4 digits


;(require 'asdf-install)
;(asdf-install:install "http://superadditive.com/software/incf-cl.tar.gz")

(require :incf-cl)
(use-package :incf-cl)

(defvar *4-digit-prime-file* "/home/mortis/personal/projects/project-euler.net/data-files/4-digit-primes.txt")

(defun get-4-digit-primes ()
  (with-open-file (in *4-digit-prime-file* :direction :input)
    (loop for line = (read-line in nil nil)
	  then (read-line in nil nil)
	  while line
	  collect line)))

(defvar *4-digit-primes* (get-4-digit-primes))
;; (setf *4-digit-primes* (mapcar #'read-from-string (get-4-digit-primes)))

;; (length *4-digit-primes*)
;; (car *4-digit-primes*)

(defun make-prime-key (prime)
  (read-from-string (format nil "~a" (sort (format nil "~d" prime) #'char<))))

;; (make-prime-key 321)

(defvar *candidate-table* nil)

(setf *candidate-table*
  (let ((ht (make-hash-table :test #'equal)))
    (loop for prime in *4-digit-primes*
	  for key = (make-prime-key prime) then (make-prime-key prime)
	  do
	  (if (gethash key ht)
	      (setf (gethash key ht)
		    (push prime (gethash key ht)))
	      (setf (gethash key ht) (list prime))))
    (loop for key being the hash-keys of ht
	  do
	  (setf (gethash key ht) (sort (gethash key ht) #'<)))
    ht))

(defun all-pairs (lst)
  (assemble (list x y)
    (<- x lst)
    (<- y (remove x lst))))

;; (all-pairs '(1 2 3))

(defun find-cands-by-diff (diff pairs)
  (loop for pair in pairs
	when (= diff (apply #'- pair))
	collect pair))

(defun list-unique (things)
  (let ((ht (make-hash-table))
	(result (list)))
    (loop for thing in things
	  do
	  (when (not (gethash thing ht))
	    (setf (gethash thing ht) t)
	    (push thing result)))
    (reverse result)))


(defun search-for-diff (key lst)
  (loop
   with pairs = (all-pairs lst)
   for pair in pairs
   for left = (first pair) then (first pair)
   for right = (second pair) then (second pair)
   for diff = (abs (- left right))
   for cands = (find-cands-by-diff diff pairs)
   then (find-cands-by-diff diff pairs)
   when (and (= diff 3330) (> (length cands) 1))
   do
   (format t "~a  ~a-~a=~a cands: ~a~&"
	   key left right diff (sort (list-unique (apply #'append cands)) #'<))))

(loop for key being the hash-keys of *candidate-table*
      for lst = (gethash key *candidate-table*)
      then (gethash key *candidate-table*)
      do
      (when (> (length lst) 2)
	(search-for-diff key lst)))



;;;;; 9629-6299=3330 cands: ((6299 2969) (9629 6299))
;; 2969 6299 9629 
;; 296962999629 




;; (defun memoize (function)
;;   (let ((res-table (make-hash-table :test #'equal :size 100000)))
;;     (lambda (&rest args)
;;       (multiple-value-bind 
;; 	    (value exists) 
;; 	  (gethash args res-table)
;; 	(when (not exists)
;; 	  (setf value (apply function args))
;; 	  (setf (gethash args res-table) value))
;; 	value))))

;; (defun inner-prime? (num)
;;   (cond ((= 1 num) nil)
;; 	((= 2 num) t)
;; 	((= 0 (mod num 2)) nil)
;; 	(t
;; 	 (loop for fact = (ceiling (sqrt num))
;; 	       then (if (oddp fact)
;; 			(- fact 2)
;; 			(- fact 1))
;; 	       while (>= fact 2)
;; 	       do
;; 	       (when (= 0 (mod num fact))
;; 		 (return-from inner-prime? nil)))
;; 	 t)))


;; (defvar *inner-prime?-mem* (memoize #'inner-prime?))

;; (defun prime? (num)
;;   (funcall *inner-prime?-mem* num))

;; (defun get-primes-between (min max &optional (start 3))
;;   (let ((res (list)))
;;     (loop for ii = start then (+ ii 2)
;; 	  while (< ii max)
;; 	  do
;; 	  (when (and (> ii min) (prime? ii))
;; 	    (push ii res)))
;;     res))

;; (defvar *four-digit-primes* (get-primes-between 999 10000))

;; ;; (length *four-digit-primes*)


;; (with-open-file (out "/home/mortis/personal/projects/project-euler.net/x.x" :direction :output :if-exists :supersede)
;;   (loop for prm in *four-digit-primes*
;; 	do
;; 	(format out "~d~&" prm)))


;; (defun char->digit (chr)
;;   (cond
;;     ((eq #\1 chr) 1)
;;     ((eq #\2 chr) 2)
;;     ((eq #\3 chr) 3)
;;     ((eq #\4 chr) 4)
;;     ((eq #\5 chr) 5)
;;     ((eq #\6 chr) 6)
;;     ((eq #\7 chr) 7)
;;     ((eq #\8 chr) 8)
;;     ((eq #\9 chr) 9)
;;     ((eq #\0 chr) 0)
;;     (t
;;      (error "unsupported char: ~a~&" chr))))

;; (defun explode-num (num)
;;   (let ((str (format nil "~d" num)))
;;     (loop for ii from 0 to (- (length str) 1)
;; 	  collect (char->digit (char str ii)))))

;; (defun list-unique (things)
;;   (let ((ht (make-hash-table))
;; 	(result (list)))
;;     (loop for thing in things
;; 	  do
;; 	  (when (not (gethash thing ht))
;; 	    (setf (gethash thing ht) t)
;; 	    (push thing result)))
;;     (reverse result)))


;; (defun find-permuted-primes (prime)
;;   (let ((digits (explode-num prime))
;; 	(res (list)))
;;     (loop 
;;      for d1 in digits ;;(remove 0 digits)
;;      for rem1 = (remove d1 digits) then (remove d1 digits)
;;      do
;;      (loop 
;;       for d2 in rem1
;;       for rem2 = (remove d2 rem1) then (remove d2 rem1)
;;       do
;;       (loop
;;        for d3 in rem2
;;        for rem3 = (remove d3 rem2) then (remove d3 rem2)
;;        do
;;        (loop
;; 	for d4 in rem3
;; 	for permuted = (read-from-string (format nil "~a~a~a~a" d1 d2 d3 d4))
;; 	then (read-from-string (format nil "~a~a~a~a" d1 d2 d3 d4))
;; 	do
;; 	(when (and (prime? permuted) (> permuted 999))
;; 	  (push permuted res))))))
;;     (list-unique (sort res #'<))))

;; ;;(find-permuted-primes 1487)
;; ;;(find-permuted-primes 1789)

;; ;;(find-permuted-primes (first *four-digit-primes*))

;; ;; (explode-num (first *four-digit-primes*))

;; ;;(find-permuted-primes 1487)

;; (defun find-with-diff-pairs (diff nums)
;;   (let ((res (list)))
;;     (loop for nn in nums
;; 	  do
;; 	  (loop for mm in nums
;; 		for dff = (abs (- mm nn)) then (abs (- mm nn))
;; 		do
;; 		(when (and
;; 		       (not (= nn mm))
;; 		       (= diff dff))
;; 		  (push (list nn mm) res))))
;;     (sort (list-unique (apply #'append res)) #'<)))

;; (defun find-with-diff (diff nums)
;;   (let ((dnums (find-with-diff-pairs diff nums)))
;;     (loop
;;      for prev = (first dnums) then (first rem-nums)
;;      for rem-nums = (rest dnums) then (rest rem-nums)
;;      do
;;      (when (not rem-nums)
;;        (return dnums))
;;      (if (not (= diff (- (first rem-nums) prev)))
;; 	 (return-from find-with-diff nil)))
;;     dnums))


;; ;; (find-with-diff 3330 (find-permuted-primes 1487))

;; (defun seen (ht value)
;;   (cond ((not (gethash value ht))
;; 	 (setf (gethash value ht) t)
;; 	 nil)
;; 	(t
;; 	 t)))

;; (defun find-next-4dig ()
;;   (let ((ht (make-hash-table :test #'equal)))
;;     (loop 
;;      for prime in (get-primes-between 999 10000)
;;      for perms = (find-permuted-primes prime) then (find-permuted-primes prime)
;;      do
;;      (loop 
;;       for perm in perms
;;       do
;;       (loop
;;        for perm2 in (remove perm perms)
;;        for diff = (abs (- perm2 perm)) then (abs (- perm2 perm))
;;        for with-diffs = (find-with-diff diff perms) then (find-with-diff diff perms)
;;        do
;;        (when (and (>= (length with-diffs) 3)
;; 		  (not (seen ht with-diffs)))
;; 	 (format t "found(~a): perm:~a perm2:~a diff:~a ~a~&" (length with-diffs) perm perm2 diff with-diffs)))))))

;; (time (find-next-4dig))

;; ;;found: perm:1789 perm2:1879 diff:90 (1789 1879 9781 9871)
;; ;; (find-permuted-primes 1487)


;; (mapcar #'prime? '(4157 4751 5147 5741))
